#!/usr/bin/env python3

from __future__ import annotations

import warnings
from typing import Any, Optional, Tuple, Union

import torch
from linear_operator import LinearOperator, operators
from linear_operator.operators.cat_linear_operator import cat as _cat

from .lazy_evaluated_kernel_tensor import LazyEvaluatedKernelTensor
from .lazy_tensor import delazify, deprecated_lazy_tensor
from .non_lazy_tensor import lazify

# We will dynamically import LazyTensor/NonLazyTensor to trigger DeprecationWarnings


_deprecated_lazy_tensors = {
    "AddedDiagLazyTensor": deprecated_lazy_tensor(operators.AddedDiagLinearOperator),
    "BatchRepeatLazyTensor": deprecated_lazy_tensor(operators.BatchRepeatLinearOperator),
    "BlockDiagLazyTensor": deprecated_lazy_tensor(operators.BlockDiagLinearOperator),
    "BlockInterleavedLazyTensor": deprecated_lazy_tensor(operators.BlockInterleavedLinearOperator),
    "BlockLazyTensor": deprecated_lazy_tensor(operators.BlockLinearOperator),
    "CatLazyTensor": deprecated_lazy_tensor(operators.CatLinearOperator),
    "CholLazyTensor": deprecated_lazy_tensor(operators.CholLinearOperator),
    "ConstantMulLazyTensor": deprecated_lazy_tensor(operators.ConstantMulLinearOperator),
    "ConstantDiagLazyTensor": deprecated_lazy_tensor(operators.ConstantDiagLinearOperator),
    "DiagLazyTensor": deprecated_lazy_tensor(operators.DiagLinearOperator),
    "IdentityLazyTensor": deprecated_lazy_tensor(operators.IdentityLinearOperator),
    "InterpolatedLazyTensor": deprecated_lazy_tensor(operators.InterpolatedLinearOperator),
    "KeOpsLazyTensor": deprecated_lazy_tensor(operators.KeOpsLinearOperator),
    "KroneckerProductAddedDiagLazyTensor": deprecated_lazy_tensor(operators.KroneckerProductAddedDiagLinearOperator),
    "KroneckerProductDiagLazyTensor": deprecated_lazy_tensor(operators.KroneckerProductDiagLinearOperator),
    "KroneckerProductLazyTensor": deprecated_lazy_tensor(operators.KroneckerProductLinearOperator),
    "KroneckerProductTriangularLazyTensor": deprecated_lazy_tensor(operators.KroneckerProductTriangularLinearOperator),
    "LowRankRootAddedDiagLazyTensor": deprecated_lazy_tensor(operators.LowRankRootAddedDiagLinearOperator),
    "LowRankRootLazyTensor": deprecated_lazy_tensor(operators.LowRankRootLinearOperator),
    "MatmulLazyTensor": deprecated_lazy_tensor(operators.MatmulLinearOperator),
    "MulLazyTensor": deprecated_lazy_tensor(operators.MulLinearOperator),
    "PsdSumLazyTensor": deprecated_lazy_tensor(operators.PsdSumLinearOperator),
    "RootLazyTensor": deprecated_lazy_tensor(operators.RootLinearOperator),
    "SumBatchLazyTensor": deprecated_lazy_tensor(operators.SumBatchLinearOperator),
    "SumKroneckerLazyTensor": deprecated_lazy_tensor(operators.SumKroneckerLinearOperator),
    "SumLazyTensor": deprecated_lazy_tensor(operators.SumLinearOperator),
    "ToeplitzLazyTensor": deprecated_lazy_tensor(operators.ToeplitzLinearOperator),
    "TriangularLazyTensor": deprecated_lazy_tensor(operators.TriangularLinearOperator),
    "ZeroLazyTensor": deprecated_lazy_tensor(operators.ZeroLinearOperator),
}


def cat(
    inputs: Tuple[Union[LinearOperator, torch.Tensor], ...], dim: int = 0, output_device: Optional[torch.device] = None
) -> Union[torch.Tensor, LinearOperator]:
    warnings.warn("gpytorch.lazy.cat is deprecated in favor of linear_operator.cat")
    return _cat(inputs, dim=dim, output_device=output_device)


__all__ = [
    "delazify",
    "lazify",
    "cat",
    "LazyEvaluatedKernelTensor",
    "LazyTensor",
]


def __getattr__(name: str) -> Any:
    warnings.warn(
        "GPyTorch will be replacing all LazyTensor functionality with the linear operator package. "
        "Replace all references to gpytorch.lazy.*LazyTensor with linear_operator.operators.*LinearOperator.",
        DeprecationWarning,
    )
    if name == "LazyTensor":
        from .lazy_tensor import LazyTensor

        return deprecated_lazy_tensor(LazyTensor)
    elif name == "NonLazyTensor":
        from .non_lazy_tensor import NonLazyTensor

        return deprecated_lazy_tensor(NonLazyTensor)
    elif name in _deprecated_lazy_tensors:
        return _deprecated_lazy_tensors[name]
    raise AttributeError(f"module gpytorch.lazy has no attribute {name}")



#!/usr/bin/env python3
import math

import torch
from linear_operator.operators import KernelLinearOperator

from ..matern_kernel import MaternKernel as GMaternKernel
from .keops_kernel import KeOpsKernel

try:
    from pykeops.torch import LazyTensor as KEOLazyTensor

    def _covar_func(x1, x2, nu=2.5, **params):
        x1_ = KEOLazyTensor(x1[..., :, None, :])
        x2_ = KEOLazyTensor(x2[..., None, :, :])

        distance = ((x1_ - x2_) ** 2).sum(-1).sqrt()
        exp_component = (-math.sqrt(nu * 2) * distance).exp()

        if nu == 0.5:
            constant_component = 1
        elif nu == 1.5:
            constant_component = (math.sqrt(3) * distance) + 1
        elif nu == 2.5:
            constant_component = (math.sqrt(5) * distance) + (1 + 5.0 / 3.0 * (distance**2))

        return constant_component * exp_component

    class MaternKernel(KeOpsKernel):
        """
        Implements the Matern kernel using KeOps as a driver for kernel matrix multiplies.

        This class can be used as a drop in replacement for :class:`gpytorch.kernels.MaternKernel` in most cases,
        and supports the same arguments.

        :param nu: (Default: 2.5) The smoothness parameter.
        :type nu: float (0.5, 1.5, or 2.5)
        :param ard_num_dims: (Default: `None`) Set this if you want a separate lengthscale for each
            input dimension. It should be `d` if x1 is a `... x n x d` matrix.
        :type ard_num_dims: int, optional
        :param batch_shape: (Default: `None`) Set this if you want a separate lengthscale for each
             batch of input data. It should be `torch.Size([b1, b2])` for a `b1 x b2 x n x m` kernel output.
        :type batch_shape: torch.Size, optional
        :param active_dims: (Default: `None`) Set this if you want to
            compute the covariance of only a few input dimensions. The ints
            corresponds to the indices of the dimensions.
        :type active_dims: Tuple(int)
        :param lengthscale_prior: (Default: `None`)
            Set this if you want to apply a prior to the lengthscale parameter.
        :type lengthscale_prior: ~gpytorch.priors.Prior, optional
        :param lengthscale_constraint: (Default: `Positive`) Set this if you want
            to apply a constraint to the lengthscale parameter.
        :type lengthscale_constraint: ~gpytorch.constraints.Interval, optional
        :param eps: (Default: 1e-6) The minimum value that the lengthscale can take (prevents divide by zero errors).
        :type eps: float, optional
        """

        has_lengthscale = True

        def __init__(self, nu=2.5, **kwargs):
            if nu not in {0.5, 1.5, 2.5}:
                raise RuntimeError("nu expected to be 0.5, 1.5, or 2.5")
            super(MaternKernel, self).__init__(**kwargs)
            self.nu = nu

        def _nonkeops_forward(self, x1, x2, diag=False, **kwargs):
            mean = x1.reshape(-1, x1.size(-1)).mean(0)[(None,) * (x1.dim() - 1)]
            x1_ = (x1 - mean) / self.lengthscale
            x2_ = (x2 - mean) / self.lengthscale

            distance = self.covar_dist(x1_, x2_, diag=diag, **kwargs)
            exp_component = torch.exp(-math.sqrt(self.nu * 2) * distance)

            if self.nu == 0.5:
                constant_component = 1
            elif self.nu == 1.5:
                constant_component = (math.sqrt(3) * distance).add(1)
            elif self.nu == 2.5:
                constant_component = (math.sqrt(5) * distance).add(1).add(5.0 / 3.0 * distance**2)
            return constant_component * exp_component

        def _keops_forward(self, x1, x2, **kwargs):
            mean = x1.reshape(-1, x1.size(-1)).mean(0)[(None,) * (x1.dim() - 1)]
            x1_ = (x1 - mean) / self.lengthscale
            x2_ = (x2 - mean) / self.lengthscale
            # return KernelLinearOperator inst only when calculating the whole covariance matrix
            return KernelLinearOperator(x1_, x2_, covar_func=_covar_func, nu=self.nu, **kwargs)

except ImportError:

    class MaternKernel(GMaternKernel):
        pass



#!/usr/bin/env python3

import warnings
from functools import wraps
from typing import Union

import torch
from linear_operator import LinearOperator, to_dense


def _add_deprecated_method(cls: type, old_method_name: str, new_method_name: str):
    method = getattr(cls, new_method_name)

    @wraps(method)
    def _deprecated_method(self, *args, **kwargs):
        warnings.warn(
            f"LazyTensor#{old_method_name} has been renamed to LinearOperator#{method.__name__}", DeprecationWarning
        )
        return method(self, *args, **kwargs)

    setattr(cls, old_method_name, _deprecated_method)


def delazify(obj: Union[LinearOperator, torch.Tensor]) -> torch.Tensor:
    warnings.warn("gpytorch.lazy.delazify is deprecated in favor of linear_operator.to_dense")
    return to_dense(obj)


def deprecated_lazy_tensor(_LinearOperatorClass: type) -> type:
    __orig_init__ = getattr(_LinearOperatorClass, "__init__")

    def __init__(self, *args, **kwargs):
        new_kwargs = dict()
        for name, val in kwargs.items():
            if "lazy_tensor" in name:
                new_name = name.replace("lazy_tensor", "linear_op")
                warnings.warn(
                    f"The kwarg {name} for {self.__class__.__name__}.__init__ is deprecated. Use "
                    f"the kwarg {new_name} instead.",
                    DeprecationWarning,
                )
                new_kwargs[new_name] = val
            else:
                new_kwargs[name] = val

        return __orig_init__(self, *args, **new_kwargs)

    def symeig(self, eigenvectors=True):
        warnings.warn(
            "LazyTensor#symeig has been renamed to LinearOperator#eigh/eigvalsh. "
            "(eigh replaces symeig(eigenvectors=True); eigvalsh replaces symeig(eigenvectors=False).)",
            DeprecationWarning,
        )
        if eigenvectors:
            return self.eigh()
        else:
            return self.eigvalsh(), None

    def __getattr__(self, name):
        if "lazy_tensor" in name:
            new_name = name.replace("lazy_tensor", "linear_op")
            if hasattr(self, new_name):
                warnings.warn(
                    f"The property {self.__class__.__name__}#{name} is deprecated. Use "
                    f"{self.__class__.__name__}#{new_name} instead."
                )
                return getattr(self, new_name)
        raise AttributeError(f"Unknown attribute {name} for {self.__class__.__name__}")

    _add_deprecated_method(_LinearOperatorClass, "_approx_diag", "_approx_diagonal")
    _add_deprecated_method(_LinearOperatorClass, "_quad_form_derivative", "_bilinear_derivative")
    _add_deprecated_method(_LinearOperatorClass, "add_diag", "add_diagonal")
    _add_deprecated_method(_LinearOperatorClass, "diag", "diagonal")
    _add_deprecated_method(_LinearOperatorClass, "evaluate", "to_dense")
    _add_deprecated_method(_LinearOperatorClass, "inv_matmul", "solve")
    setattr(_LinearOperatorClass, "symeig", symeig)
    setattr(_LinearOperatorClass, "__getattr__", __getattr__)
    setattr(_LinearOperatorClass, "__init__", __init__)
    return _LinearOperatorClass


LazyTensor = deprecated_lazy_tensor(LinearOperator)



#!/usr/bin/env python3

import warnings
from typing import Union

import torch
from linear_operator import LinearOperator, operators, to_linear_operator

from .lazy_tensor import deprecated_lazy_tensor


def lazify(obj: Union[LinearOperator, torch.Tensor]) -> operators.DenseLinearOperator:
    warnings.warn("gpytorch.lazy.lazify is deprecated in favor of linear_operator.to_linear_operator")
    return to_linear_operator(obj)


NonLazyTensor = deprecated_lazy_tensor(operators.DenseLinearOperator)



#!/usr/bin/env python3

import unittest
import warnings

import linear_operator
import torch

import gpytorch
from gpytorch.test.base_test_case import BaseTestCase

lazy_tensor_map = {
    "AddedDiagLazyTensor": linear_operator.operators.AddedDiagLinearOperator,
    "BatchRepeatLazyTensor": linear_operator.operators.BatchRepeatLinearOperator,
    "BlockDiagLazyTensor": linear_operator.operators.BlockDiagLinearOperator,
    "BlockInterleavedLazyTensor": linear_operator.operators.BlockInterleavedLinearOperator,
    "BlockLazyTensor": linear_operator.operators.BlockLinearOperator,
    "CatLazyTensor": linear_operator.operators.CatLinearOperator,
    "CholLazyTensor": linear_operator.operators.CholLinearOperator,
    "ConstantMulLazyTensor": linear_operator.operators.ConstantMulLinearOperator,
    "ConstantDiagLazyTensor": linear_operator.operators.ConstantDiagLinearOperator,
    "DiagLazyTensor": linear_operator.operators.DiagLinearOperator,
    "IdentityLazyTensor": linear_operator.operators.IdentityLinearOperator,
    "InterpolatedLazyTensor": linear_operator.operators.InterpolatedLinearOperator,
    "KeOpsLazyTensor": linear_operator.operators.KeOpsLinearOperator,
    "KroneckerProductAddedDiagLazyTensor": linear_operator.operators.KroneckerProductAddedDiagLinearOperator,
    "KroneckerProductDiagLazyTensor": linear_operator.operators.KroneckerProductDiagLinearOperator,
    "KroneckerProductLazyTensor": linear_operator.operators.KroneckerProductLinearOperator,
    "KroneckerProductTriangularLazyTensor": linear_operator.operators.KroneckerProductTriangularLinearOperator,
    "LazyTensor": linear_operator.operators.LinearOperator,
    "LowRankRootAddedDiagLazyTensor": linear_operator.operators.LowRankRootAddedDiagLinearOperator,
    "LowRankRootLazyTensor": linear_operator.operators.LowRankRootLinearOperator,
    "MatmulLazyTensor": linear_operator.operators.MatmulLinearOperator,
    "MulLazyTensor": linear_operator.operators.MulLinearOperator,
    "NonLazyTensor": linear_operator.operators.DenseLinearOperator,
    "PsdSumLazyTensor": linear_operator.operators.PsdSumLinearOperator,
    "RootLazyTensor": linear_operator.operators.RootLinearOperator,
    "SumBatchLazyTensor": linear_operator.operators.SumBatchLinearOperator,
    "SumKroneckerLazyTensor": linear_operator.operators.SumKroneckerLinearOperator,
    "SumLazyTensor": linear_operator.operators.SumLinearOperator,
    "ToeplitzLazyTensor": linear_operator.operators.ToeplitzLinearOperator,
    "TriangularLazyTensor": linear_operator.operators.TriangularLinearOperator,
    "ZeroLazyTensor": linear_operator.operators.ZeroLinearOperator,
}


class LazyTensorImport(BaseTestCase, unittest.TestCase):
    def test_deprecated_imports(self):
        for lt_class_name, linear_op_class in lazy_tensor_map.items():
            with warnings.catch_warnings(record=True) as ws:
                lt_class = getattr(gpytorch.lazy, lt_class_name)
                self.assertEqual(lt_class, linear_op_class)
                self.assertTrue(len(ws) == 1)
                self.assertTrue(issubclass(ws[0].category, DeprecationWarning))

    def test_deprecated_methods(self):
        with warnings.catch_warnings(record=True) as ws:
            mat = torch.randn(5, 5)
            mat = mat @ mat.mT
            nlt = gpytorch.lazy.NonLazyTensor(mat)
            dlo = linear_operator.operators.DenseLinearOperator(mat)
            self.assertTrue(len(ws) == 1)
            self.assertTrue(issubclass(ws[0].category, DeprecationWarning))

        with warnings.catch_warnings(record=True) as ws:
            diag = torch.randn(5)
            self.assertAllClose(nlt.add_diag(diag).to_dense(), dlo.add_diagonal(diag).to_dense())
            self.assertTrue(len(ws) == 1)
            self.assertTrue(issubclass(ws[0].category, DeprecationWarning))

        with warnings.catch_warnings(record=True) as ws:
            self.assertAllClose(nlt.diag(), dlo.diag())
            self.assertTrue(len(ws) == 1)
            self.assertTrue(issubclass(ws[0].category, DeprecationWarning))

        with warnings.catch_warnings(record=True) as ws:
            self.assertAllClose(nlt.evaluate(), dlo.to_dense())
            self.assertTrue(len(ws) == 1)
            self.assertTrue(issubclass(ws[0].category, DeprecationWarning))

        with warnings.catch_warnings(record=True) as ws:
            rhs = torch.randn(5, 2)
            self.assertAllClose(nlt.inv_matmul(rhs), dlo.solve(rhs))
            self.assertTrue(len(ws) == 1)
            self.assertTrue(issubclass(ws[0].category, DeprecationWarning))

        with warnings.catch_warnings(record=True) as ws:
            self.assertAllClose(nlt.symeig(eigenvectors=True)[0], torch.linalg.eigh(dlo)[0])
            self.assertAllClose(nlt.symeig(eigenvectors=True)[1].to_dense(), torch.linalg.eigh(dlo)[1].to_dense())
            self.assertTrue(len(ws) == 1)
            self.assertTrue(issubclass(ws[0].category, DeprecationWarning))

        with warnings.catch_warnings(record=True) as ws:
            self.assertAllClose(nlt.symeig(eigenvectors=False)[0], torch.linalg.eigvalsh(dlo))
            self.assertTrue(len(ws) == 1)
            self.assertTrue(issubclass(ws[0].category, DeprecationWarning))

        with warnings.catch_warnings(record=True) as ws:
            a = torch.randn(5, 2)
            b = torch.randn(5, 2)
            self.assertAllClose(nlt._quad_form_derivative(a, b)[0], dlo._bilinear_derivative(a, b)[0])
            self.assertTrue(len(ws) == 1)
            self.assertTrue(issubclass(ws[0].category, DeprecationWarning))



#!/usr/bin/env python3

import unittest

import torch

import gpytorch
from gpytorch.test.variational_test_case import VariationalTestCase


class TestVNNGP(VariationalTestCase, unittest.TestCase):
    @property
    def batch_shape(self):
        return torch.Size([])

    @property
    def distribution_cls(self):
        return gpytorch.variational.MeanFieldVariationalDistribution

    @property
    def mll_cls(self):
        return gpytorch.mlls.VariationalELBO

    @property
    def strategy_cls(self):
        return gpytorch.variational.NNVariationalStrategy

    @property
    def likelihood_cls(self):
        return gpytorch.likelihoods.GaussianLikelihood

    @property
    def event_shape(self):
        return torch.Size([32])

    def _make_model_and_likelihood(
        self,
        num_inducing=32,
        batch_shape=torch.Size([]),
        inducing_batch_shape=torch.Size([]),
        strategy_cls=gpytorch.variational.NNVariationalStrategy,
        distribution_cls=gpytorch.variational.MeanFieldVariationalDistribution,
        constant_mean=True,
    ):
        class _VNNGPRegressionModel(gpytorch.models.GP):
            def __init__(self, inducing_points, k, training_batch_size):
                super(_VNNGPRegressionModel, self).__init__()

                variational_distribution = distribution_cls(num_inducing, batch_shape=batch_shape)

                self.variational_strategy = strategy_cls(
                    self, inducing_points, variational_distribution, k=k, training_batch_size=training_batch_size
                )

                if constant_mean:
                    self.mean_module = gpytorch.means.ConstantMean()
                    self.mean_module.initialize(constant=1.0)
                else:
                    self.mean_module = gpytorch.means.ZeroMean()

                self.covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())

            def forward(self, x):
                mean_x = self.mean_module(x)
                covar_x = self.covar_module(x)
                return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)

            def __call__(self, x, prior=False, **kwargs):
                if x is not None:
                    if x.dim() == 1:
                        x = x.unsqueeze(-1)
                return self.variational_strategy(x=x, prior=False, **kwargs)

        k = 3
        d = 2
        training_batch_size = 4
        inducing_points = torch.randn(*inducing_batch_shape, num_inducing, d)
        return _VNNGPRegressionModel(inducing_points, k, training_batch_size), self.likelihood_cls()

    def _training_iter(
        self,
        model,
        likelihood,
        batch_shape=torch.Size([]),
        mll_cls=gpytorch.mlls.VariationalELBO,
        cuda=False,
    ):
        # We cannot inheret the superclass method
        # Because it sets the training data to be the inducing points

        train_x = model.variational_strategy.inducing_points
        train_y = torch.randn(train_x.shape[:-1])
        mll = mll_cls(likelihood, model, num_data=train_x.size(-2))
        if cuda:
            train_x = train_x.cuda()
            train_y = train_y.cuda()
            model = model.cuda()
            likelihood = likelihood.cuda()

        # Single optimization iteration
        model.train()
        likelihood.train()
        output = model(train_x)
        loss = -mll(output, train_y)
        loss.sum().backward()

        # Make sure we have gradients for all parameters
        for _, param in model.named_parameters():
            self.assertTrue(param.grad is not None)
            self.assertGreater(param.grad.norm().item(), 0)
        for _, param in likelihood.named_parameters():
            self.assertTrue(param.grad is not None)
            self.assertGreater(param.grad.norm().item(), 0)

        return output, loss

    def _eval_iter(self, model, cuda=False):
        inducing_batch_shape = model.variational_strategy.inducing_points.shape[:-2]
        test_x = torch.randn(*inducing_batch_shape, 32, 2).clamp(-2.5, 2.5)
        if cuda:
            test_x = test_x.cuda()
            model = model.cuda()

        # Single optimization iteration
        model.eval()
        with torch.no_grad():
            output = model(test_x)

        return output

    def test_training_iteration(
        self,
        data_batch_shape=None,
        inducing_batch_shape=None,
        model_batch_shape=None,
        expected_batch_shape=None,
        constant_mean=True,
    ):
        # We cannot inheret the superclass method
        # Because it expects `variational_params_intialized` to be set to 0

        # Batch shapes
        model_batch_shape = model_batch_shape if model_batch_shape is not None else self.batch_shape
        data_batch_shape = data_batch_shape if data_batch_shape is not None else self.batch_shape
        inducing_batch_shape = inducing_batch_shape if inducing_batch_shape is not None else self.batch_shape
        expected_batch_shape = expected_batch_shape if expected_batch_shape is not None else self.batch_shape

        # Make model and likelihood
        model, likelihood = self._make_model_and_likelihood(
            batch_shape=model_batch_shape,
            inducing_batch_shape=inducing_batch_shape,
            distribution_cls=self.distribution_cls,
            strategy_cls=self.strategy_cls,
            constant_mean=constant_mean,
        )

        # Do forward pass
        # Iter 1
        self._training_iter(
            model,
            likelihood,
            data_batch_shape,
            mll_cls=self.mll_cls,
            cuda=self.cuda,
        )
        # Iter 2
        output, loss = self._training_iter(
            model,
            likelihood,
            data_batch_shape,
            mll_cls=self.mll_cls,
            cuda=self.cuda,
        )
        self.assertEqual(output.batch_shape, expected_batch_shape)
        self.assertEqual(output.event_shape, self.event_shape)
        self.assertEqual(loss.shape, expected_batch_shape)

    def test_training_iteration_batch_inducing(self):
        # We need different batch sizes than the superclass
        return self.test_training_iteration(
            model_batch_shape=(torch.Size([3]) + self.batch_shape),
            inducing_batch_shape=(torch.Size([3]) + self.batch_shape),
            expected_batch_shape=(torch.Size([3]) + self.batch_shape),
        )

    def test_training_iteration_batch_data(self):
        # We need different batch sizes than the superclass
        return self.test_training_iteration(
            model_batch_shape=self.batch_shape,
            inducing_batch_shape=self.batch_shape,
            expected_batch_shape=(self.batch_shape),
        )

    def test_training_iteration_batch_model(self):
        # We need different batch sizes than the superclass
        return self.test_training_iteration(
            model_batch_shape=(torch.Size([3]) + self.batch_shape),
            inducing_batch_shape=self.batch_shape,
            expected_batch_shape=(torch.Size([3]) + self.batch_shape),
        )

    def test_training_all_batch_zero_mean(self):
        # We need different batch sizes than the superclass
        return self.test_training_iteration(
            model_batch_shape=(torch.Size([3, 4]) + self.batch_shape),
            inducing_batch_shape=(torch.Size([3, 1]) + self.batch_shape),
            expected_batch_shape=(torch.Size([3, 4]) + self.batch_shape),
            constant_mean=False,
        )

    def test_eval_iteration(
        self,
        inducing_batch_shape=None,
        model_batch_shape=None,
        expected_batch_shape=None,
    ):
        # Batch shapes
        model_batch_shape = model_batch_shape if model_batch_shape is not None else self.batch_shape
        inducing_batch_shape = inducing_batch_shape if inducing_batch_shape is not None else self.batch_shape
        expected_batch_shape = expected_batch_shape if expected_batch_shape is not None else self.batch_shape

        # Make model and likelihood
        model, likelihood = self._make_model_and_likelihood(
            batch_shape=model_batch_shape,
            inducing_batch_shape=inducing_batch_shape,
        )

        # Do one forward pass
        self._training_iter(model, likelihood, mll_cls=self.mll_cls, cuda=self.cuda)

        # Now do evaluation
        # Iter 1
        _ = self._eval_iter(model, cuda=self.cuda)
        output = self._eval_iter(model, cuda=self.cuda)
        self.assertEqual(output.batch_shape, expected_batch_shape)
        self.assertEqual(output.event_shape, self.event_shape)

    def test_eval_smaller_pred_batch(self):
        # We need different batch sizes than the superclass
        return self.test_eval_iteration(
            model_batch_shape=(torch.Size([3, 4]) + self.batch_shape),
            inducing_batch_shape=(torch.Size([3, 1]) + self.batch_shape),
            expected_batch_shape=(torch.Size([3, 4]) + self.batch_shape),
        )

    def test_eval_larger_pred_batch(self):
        # We need different batch sizes than the superclass
        return self.test_eval_iteration(
            model_batch_shape=(torch.Size([4]) + self.batch_shape),
            inducing_batch_shape=(self.batch_shape),
            expected_batch_shape=(torch.Size([4]) + self.batch_shape),
        )

    def test_fantasy_call(self, *args, **kwargs):
        with self.assertRaises(AttributeError):
            super().test_fantasy_call(*args, **kwargs)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

# from linear_operator.operators import KeOpsLinearOperator
from linear_operator.operators import KernelLinearOperator

from ..rbf_kernel import postprocess_rbf, RBFKernel as GRBFKernel
from .keops_kernel import KeOpsKernel

try:
    from pykeops.torch import LazyTensor as KEOLazyTensor

    def _covar_func(x1, x2, **kwargs):
        x1_ = KEOLazyTensor(x1[..., :, None, :])
        x2_ = KEOLazyTensor(x2[..., None, :, :])
        K = (-((x1_ - x2_) ** 2).sum(-1) / 2).exp()
        return K

    class RBFKernel(KeOpsKernel):
        r"""
        Implements the RBF kernel using KeOps as a driver for kernel matrix multiplies.

        This class can be used as a drop in replacement for :class:`gpytorch.kernels.RBFKernel` in most cases,
        and supports the same arguments.

        :param ard_num_dims: Set this if you want a separate lengthscale for each input
            dimension. It should be `d` if x1 is a `n x d` matrix. (Default: `None`.)
        :param batch_shape: Set this if you want a separate lengthscale for each batch of input
            data. It should be :math:`B_1 \times \ldots \times B_k` if :math:`\mathbf x1` is
            a :math:`B_1 \times \ldots \times B_k \times N \times D` tensor.
        :param active_dims: Set this if you want to compute the covariance of only
            a few input dimensions. The ints corresponds to the indices of the
            dimensions. (Default: `None`.)
        :param lengthscale_prior: Set this if you want to apply a prior to the
            lengthscale parameter. (Default: `None`)
        :param lengthscale_constraint: Set this if you want to apply a constraint
            to the lengthscale parameter. (Default: `Positive`.)
        :param eps: The minimum value that the lengthscale can take (prevents
            divide by zero errors). (Default: `1e-6`.)

        :ivar torch.Tensor lengthscale: The lengthscale parameter. Size/shape of parameter depends on the
            ard_num_dims and batch_shape arguments.
        """

        has_lengthscale = True

        def _nonkeops_forward(self, x1, x2, diag=False, **kwargs):
            x1_ = x1 / self.lengthscale
            x2_ = x2 / self.lengthscale
            return postprocess_rbf(self.covar_dist(x1_, x2_, square_dist=True, diag=diag, **kwargs))

        def _keops_forward(self, x1, x2, **kwargs):
            x1_ = x1 / self.lengthscale
            x2_ = x2 / self.lengthscale
            # return KernelLinearOperator inst only when calculating the whole covariance matrix
            return KernelLinearOperator(x1_, x2_, covar_func=_covar_func, **kwargs)

except ImportError:

    class RBFKernel(GRBFKernel):
        pass



from abc import abstractmethod
from typing import Any

import torch
from torch import Tensor

from ... import settings
from ..kernel import Kernel

try:
    import pykeops  # noqa F401

    class KeOpsKernel(Kernel):
        @abstractmethod
        def _nonkeops_forward(self, x1: Tensor, x2: Tensor, diag: bool = False, **kwargs: Any):
            r"""
            Computes the covariance matrix (or diagonal) without using KeOps.
            This function must implement both the diag=True and diag=False options.
            """
            raise NotImplementedError

        @abstractmethod
        def _keops_forward(self, x1: Tensor, x2: Tensor, **kwargs: Any):
            r"""
            Computes the covariance matrix with KeOps.
            This function only implements the diag=False option, and no diag keyword should be passed in.
            """
            raise NotImplementedError

        def forward(self, x1: Tensor, x2: Tensor, diag: bool = False, **kwargs: Any):
            if diag:
                return self._nonkeops_forward(x1, x2, diag=True, **kwargs)
            elif x1.size(-2) < settings.max_cholesky_size.value() or x2.size(-2) < settings.max_cholesky_size.value():
                return self._nonkeops_forward(x1, x2, diag=False, **kwargs)
            else:
                return self._keops_forward(x1, x2, **kwargs)

        def __call__(self, *args: Any, **kwargs: Any):
            # Hotfix for zero gradients. See https://github.com/cornellius-gp/gpytorch/issues/1543
            args = [arg.contiguous() if torch.is_tensor(arg) else arg for arg in args]
            kwargs = {k: v.contiguous() if torch.is_tensor(v) else v for k, v in kwargs.items()}
            return super().__call__(*args, **kwargs)

except ImportError:

    class KeOpsKernel(Kernel):
        def __init__(self, *args: Any, **kwargs: Any):
            raise RuntimeError("You must have KeOps installed to use a KeOpsKernel")



#!/usr/bin/env python3

import math

from linear_operator.operators import KernelLinearOperator

from ..periodic_kernel import PeriodicKernel as GPeriodicKernel
from .keops_kernel import KeOpsKernel

# from ...kernels import PeriodicKernel gives a cyclic import

try:
    from pykeops.torch import LazyTensor as KEOLazyTensor

    def _covar_func(x1, x2, lengthscale, **kwargs):
        # symbolic array of shape ..., ndatax1_ x 1 x ndim
        x1_ = KEOLazyTensor(x1[..., :, None, :])
        # symbolic array of shape ..., 1 x ndatax2_ x ndim
        x2_ = KEOLazyTensor(x2[..., None, :, :])
        lengthscale = lengthscale[..., None, None, 0, :]  # 1 x 1 x ndim
        # do not use .power(2.0) as it gives NaN values on cuda
        # seems related to https://github.com/getkeops/keops/issues/112
        K = ((((x1_ - x2_).abs().sin()) ** 2) * (-2.0 / lengthscale)).sum(-1).exp()
        return K

    # subclass from original periodic kernel to reduce code duplication
    class PeriodicKernel(KeOpsKernel, GPeriodicKernel):
        """
        Implements the Periodic Kernel using KeOps as a driver for kernel matrix multiplies.

        This class can be used as a drop in replacement for :class:`gpytorch.kernels.PeriodicKernel` in most cases,
        and supports the same arguments.

        :param ard_num_dims: (Default: `None`) Set this if you want a separate lengthscale for each
            input dimension. It should be `d` if x1 is a `... x n x d` matrix.
        :type ard_num_dims: int, optional
        :param batch_shape: (Default: `None`) Set this if you want a separate lengthscale for each
             batch of input data. It should be `torch.Size([b1, b2])` for a `b1 x b2 x n x m` kernel output.
        :type batch_shape: torch.Size, optional
        :param active_dims: (Default: `None`) Set this if you want to
            compute the covariance of only a few input dimensions. The ints
            corresponds to the indices of the dimensions.
        :type active_dims: Tuple(int)
        :param period_length_prior: (Default: `None`)
            Set this if you want to apply a prior to the period length parameter.
        :type period_length_prior: ~gpytorch.priors.Prior, optional
        :param period_length_constraint: (Default: `Positive`) Set this if you want
            to apply a constraint to the period length parameter.
        :type period_length_constraint: ~gpytorch.constraints.Interval, optional
        :param lengthscale_prior: (Default: `None`)
            Set this if you want to apply a prior to the lengthscale parameter.
        :type lengthscale_prior: ~gpytorch.priors.Prior, optional
        :param lengthscale_constraint: (Default: `Positive`) Set this if you want
            to apply a constraint to the lengthscale parameter.
        :type lengthscale_constraint: ~gpytorch.constraints.Interval, optional
        :param eps: (Default: 1e-6) The minimum value that the lengthscale can take (prevents divide by zero errors).
        :type eps: float, optional

        :var torch.Tensor period_length: The period length parameter. Size/shape of parameter depends on the
            ard_num_dims and batch_shape arguments.
        """

        has_lengthscale = True

        # code from the already-implemented Periodic Kernel
        def _nonkeops_forward(self, x1, x2, diag=False, **kwargs):
            x1_ = x1.div(self.period_length / math.pi)
            x2_ = x2.div(self.period_length / math.pi)

            # We are automatically overriding last_dim_is_batch here so that we can manually sum over dimensions.
            diff = self.covar_dist(x1_, x2_, diag=diag, last_dim_is_batch=True)

            if diag:
                lengthscale = self.lengthscale[..., 0, :, None]
            else:
                lengthscale = self.lengthscale[..., 0, :, None, None]

            exp_term = diff.sin().pow(2.0).div(lengthscale).mul(-2.0)
            exp_term = exp_term.sum(dim=(-2 if diag else -3))

            return exp_term.exp()

        def _keops_forward(self, x1, x2, **kwargs):
            x1_ = x1.div(self.period_length / math.pi)
            x2_ = x2.div(self.period_length / math.pi)
            # return KernelLinearOperator inst only when calculating the whole covariance matrix
            # pass any parameters which are used inside _covar_func as *args to get gradients computed for them
            return KernelLinearOperator(x1_, x2_, lengthscale=self.lengthscale, covar_func=_covar_func, **kwargs)

except ImportError:

    class PeriodicKernel(GPeriodicKernel):
        pass
